# ***************************************************************************
#
# BatterySensor - gDesklet sensor for laptop battery data.
#
# Author: Nicholas Rahn <nick at mobiledude dot com>
# Creation Date: 4 Jan. 2004
#
# Version: 0.3
#
# See README for history.
#
# ***************************************************************************

from sensor.Sensor import Sensor
from utils.datatypes import *
from utils import i18n

import re
import os

#
# Battery sensor class for gDesklets.
#
# INPUTS: Current implementation works with:
#   APM (2.4 and 2.6 kernels)
#   ACPI (BAT0 or BAT1 or both), (ac_adapter at AC or ACAD)
#
# OUTPUTS:
#   status (percentage or "AC")
#   ac ("on-line" or "off-line")
#   percent (percentage battery remaining)
#   level ("full", "ok", "warn", alert" or "critical")
#   icon (icon to use, in the form gfx/<name>.png)
#   ac_in (boolean, whether or not ac is plugged in)
#
class BatterySensor( Sensor ):

    RE_ACADAPTER = re.compile( "state:[\s]*on-line.*" )
    RE_BATPRESENT = re.compile( "present:[\s]*yes.*",
                               re.DOTALL );
    RE_CURRBATT = re.compile( ".*remaining capacity:[\s]*(?P<curr>[\d]+) .*",
                              re.DOTALL )
    RE_FULLBATT = re.compile( ".*last full capacity:[\s]*(?P<full>[\d]+) .*",
                              re.DOTALL )
    
    ACPI_BAT_BASE = "/proc/acpi/battery/"
    ACPI_INFO = ""
    ACPI_STATE = ""
    ACPI_AC_BASE = "/proc/acpi/ac_adapter/"
    ACPI_AC_LOC = ""
    
    APM_PROC = "/proc/apm"
    # Regular expression for apm status line.  This works on both 2.4 and 2.6
    # kernels because of the [\w]* bit on the end of the first grouping.
    RE_APMPROC = re.compile( "(?P<junk1>[\w\.]+) (?P<junk2>[\w\.]+) (?P<apmflags>[\w]+) (?P<acstat>[\w]+) (?P<batstat>[\w]+) (?P<batflags>[\w]+) (?P<s1>[\w\?\-\%]+) (?P<s2>[\w\?\-\%]+) (?P<s3>[\w\?\-\%]+)" )
    
    FULL_LEVEL     = "full"
    OK_LEVEL       = "ok"
    WARN_LEVEL     = "warn"
    ALERT_LEVEL    = "alert"
    CRITICAL_LEVEL = "critical"

    OK_PERCENT       = 100
    WARN_PERCENT     = 40
    ALERT_PERCENT    = 25
    CRITICAL_PERCENT = 15
    
    DEFAULT_DELAY = 2000
    
    def __init__( self,
                  warn_below = 40, alert_below = 25, critical_below = 15 ):
        """
        Initialise: The parameters give the percentages below which the
        warning, alert or critical battery warnings should be shown.
        """

        global _; _ = i18n.Translator("battery-sensor")

        # Instance variable for general battery status.
        self.__status = "AC"
        # Instance variable for AC status.
        self.__ac = "on-line"
        self.__ac_in = 1
        # Instance variable for battery percentage.
        self.__percent = self.OK_PERCENT
        # Instance variable for battery level.
        self.__color = self.FULL_LEVEL
        # Instance variable for full battery level.
        self.__fullcapacity = 0
        # Instance variable for type of battery (ACPI, APM)
        self.__type = "acpi"
        
        # Call the constructor of the parent class.
        Sensor.__init__( self )

        self._set_config_type( "warn-below",
                               TYPE_INT, int(warn_below) )
        self._set_config_type( "alert-below",
                               TYPE_INT, int(alert_below) )
        self._set_config_type( "critical-below",
                               TYPE_INT, int(critical_below) )
        self._set_config_type( "updatedelay",
                               TYPE_INT, self.DEFAULT_DELAY )

        # Auto detect the type of battery.
        if ( os.access( "/proc/acpi/battery/BAT0/info", os.F_OK ) or
             os.access( "/proc/acpi/battery/BAT1/info", os.F_OK ) ):

            self.__type = "acpi"
            bnum = self.getACPIBatNum()
            if ( bnum != "all" ):
                self.ACPI_INFO = [ self.ACPI_BAT_BASE + \
                                   "BAT" + bnum + "/info" ]
                self.ACPI_STATE = [ self.ACPI_BAT_BASE + \
                                    "BAT" + bnum + "/state" ]
            else:
                self.ACPI_INFO = [ self.ACPI_BAT_BASE + "BAT0/info",
                                   self.ACPI_BAT_BASE + "BAT1/info" ]
                self.ACPI_STATE = [ self.ACPI_BAT_BASE + "BAT0/state",
                                    self.ACPI_BAT_BASE + "BAT1/state" ]
            
        elif ( os.access( self.APM_PROC, os.F_OK ) ):
            self.__type = "apm"
        else:
            self.__type = "unknown"

        # Initialize the battery type.
        if ( self.__type == "acpi" ):
            self.initACPI()
        elif ( self.__type == "apm" ):
            self.initAPM()
    
    def get_configurator( self ):
        "Get the configurator."
        configurator = self._new_configurator()
        configurator.set_name( _("Battery") )
        configurator.add_title( _("Battery Levels") )
        configurator.add_spin( _("Warn below:"), "warn-below", 
                               _("Show warning if battery below this level "
                                 "(0-100)"), 0, 100 )
        configurator.add_spin( _("Alert below:"), "alert-below", 
                               _("Show alert if battery below this level "
                                 "(0-100)"), 0, 100 )
        configurator.add_spin( _("Critical below:"), "critical-below", 
                               _("Show critical warning if battery below "
                                 "this level (0-100)"), 0, 100 )
        
#         configurator.add_title( _("Other Configuration") )
#         configurator.add_spin( _("Update Delay (milliseconds):"),
#                                "updatedelay",
#                                _("The number of milliseconds between "
#                                  "status updates."),
#                                500, 10000 )
        return configurator
    

    # Initialize APM.
    def initAPM( self ):
        self.__status = "not supported"
        self.__fullcapacity = 0

        # Run the timer once every 2 seconds.
        self._add_timer( self.DEFAULT_DELAY, self.setAPMBatteryOutput )
        return

    # Set the "output" variables for the sensor using APM battery.
    def setAPMBatteryOutput( self ):
        try:
            f = open( self.APM_PROC, "r" )
            data = f.read()
            f.close()
        except:
            # Can't get AC state! Probably not using APM.
            return 0
        
        self.updateAPMBatteryState( data )
        
        output = self._new_output()
        output.set("status", self.__status)
        output.set("ac", self.__ac)
        output.set("ac_in", self.__ac_in)
        output.set("percent", self.__percent)
        output.set("level", self.__level)
        output.set("icon", "gfx/" + self.__level + ".png" )
        self._send_output( output )
        return 1

    # Update the member data with the current state of the APM battery.
    def updateAPMBatteryState( self, data ):
        #if ( self.RE_APMPROC.match( data ) ):
        acstat = self.RE_APMPROC.search( data ).group( "acstat" )
        s1 = self.RE_APMPROC.search( data ).group( "s1" )
            
        if ( acstat == "0x01" ):
            self.__ac = "on-line"
            self.__ac_in = 1
            self.__status = "AC"
            self.__percent = int(s1[0:(len(s1)-1)])
            self.__level = self.calcLevel( self.__percent )
        else:
            self.__ac = "off-line"
            self.__ac_in = 2
            self.__status = s1
            self.__percent = int(s1[0:(len(s1)-1)])
            self.__level = self.calcLevel( self.__percent )


    def getACPICapacity( self, info ):
        cap = 0
        try:
            f = open( info, "r" )
            data = f.read()
            cap = float(self.RE_FULLBATT.
                        search( data ).group( "full" ))
            f.close()
        except:
            # Can't get battery info! Probably not using ACPI.
            cap = 0;

        return cap

    def getACPIRemainingCapacity( self, state ):
        try:
            f = open( state, "r" )
            data = f.read()
            current = self.RE_CURRBATT.search( data ).group( "curr" )
            f.close()
        except:
            # Can't get battery state! Probably not using ACPI!
            return 0

        # Set the percentage of battery left.
        return int(current)

        
    # Initialize ACPI2 (2 batteries).
    def initACPI( self ):
        for info in self.ACPI_INFO:
            # Get the full capacity of the battery.  We do this here to avoid
            # the performance hit we get every time we read /BATn/info.
            self.__fullcapacity += self.getACPICapacity( info )

        # Set the ac_adapter location.
        self.ACPI_AC_LOC = self.ACPI_AC_BASE + \
                           self.getACPIACAdapterName() + \
                           "/state"

        # Run the timer once every 2 seconds.
        self._add_timer( self.DEFAULT_DELAY, self.setACPIBatteryOutput )


    # Set the "output" variables for the sensor using ACPI battery.
    def setACPIBatteryOutput( self ):
        try:
            f = open( self.ACPI_AC_LOC, "r" )
            data = f.read()
            f.close()
        except:
            # Can't get AC state! Probably not using ACPI.
            return 0
        
        if ( self.RE_ACADAPTER.match( data ) ):
            self.__ac = "on-line"
            self.__ac_in = 1
            self.updateACPIBatteryState()
            self.__status = "AC"
        else:
            self.__ac = "off-line"
            self.__ac_in = 0
            self.updateACPIBatteryState()
            self.__status = str(self.__percent) + "%"
            
        output = self._new_output()
        output.set("status", self.__status)
        output.set("ac", self.__ac)
        output.set("ac_in", self.__ac_in)
        output.set("percent", self.__percent)
        output.set("level", self.__level)
        output.set("icon", "gfx/" + self.__level + ".png")
        self._send_output( output )
        #print str( self.__status ) + "," + str( self.__ac ) + "," + \
        #       str( self.__ac_in ) + "," + \
        #       str( self.__percent ) + "," + str( self.__level ) + "," + \
        #       "gfx/" + self.__level + ".png," + \
        #       str( self._get_config( "updatedelay" ) )
        return 1

    # Update the member data with the current state of the ACPI battery.
    def updateACPIBatteryState( self ):
        current = 0
        for state in self.ACPI_STATE:
            current += self.getACPIRemainingCapacity( state )
            
        self.__percent = int(current/self.__fullcapacity * 100)
        # Correct for any differences between "design capacity", "last full
        # capacity" and "remaining capacity".
        if ( self.__percent > self.OK_PERCENT ):
            self.__percent = self.OK_PERCENT
        # Set the level code.
        self.__level = self.calcLevel( self.__percent )


    # Get the ACPI battery number.
    def getACPIBatNum( self ):
        ret = ""
        
        test = self.testACPILocation( "/proc/acpi/battery/BAT0/info" )
        if ( test != None ):
            if ( self.RE_BATPRESENT.match( test ) ):
                ret = ret + "0"

        test = self.testACPILocation( "/proc/acpi/battery/BAT1/info" )
        if ( test != None ):
            if ( self.RE_BATPRESENT.match( test ) ):
                ret = ret + "1"

        if ( len( ret ) > 1 ):
            return "all"
        elif ( len( ret ) == 1 ):
            return ret
        else:
            return "0"

        
    def getACPIACAdapterName( self ):
        test = self.testACPILocation( "/proc/acpi/ac_adapter/ACAD/state" )
        if ( test != None ):
            return "ACAD"
        test = self.testACPILocation( "/proc/acpi/ac_adapter/AC/state" )
        if ( test != None ):
            return "AC"
        # Defaults to AC.
        return "AC"

    def testACPILocation( self, loc ):
        try:
            f = open( loc, "r" )
            data = f.read()
            f.close()
        except:
            # BATX does not exist.
            return None
        return data
    
    def calcLevel( self, per ):
        # Calculate the level code from the percent.
         
        warn_below = self._get_config( "warn-below" )
        alert_below = self._get_config( "alert-below" )
        critical_below = self._get_config( "critical-below" )

        if ( per >= 100 ):
            return self.FULL_LEVEL
        elif ( per > warn_below ):
            return self.OK_LEVEL
        elif ( per > alert_below ):
            return self.WARN_LEVEL
        elif ( per > critical_below ):
            return self.ALERT_LEVEL
        else:
            return self.CRITICAL_LEVEL
            

def new_sensor( args ): return apply( BatterySensor, args )
