# coding=UTF-8

from sensor.Sensor import Sensor
from utils.datatypes import *

from random import *

import time
import libdesklets

#
# Sensor for BubbleFishyMon
# Version 0.1 (16/04/2004)
#
# By Vassili Geronimos (mr_spot@optusnet.com.au)
#
# Based on the BubbleFishyMon dockapp
#


from transform import *
from colour import *

MAX_FISHIES = 50
from Fishy import Fishy
from Water import Water
from Duck import Duck

class BubbleFishyMon(Sensor):


	def __init__(self):
		Sensor.__init__(self)
		
		self.water = Water()
		self.water.height = CANVAS_HEIGHT

		self.duck = Duck()
		
		self.tx_fishies = []
		self.rx_fishies = []

		# These "scores' determine how many fishies to spawn, and what
		# speed they swim at
		self.tx_points = 0
		self.rx_points = 0

		self.cpu = 0
		self.swap_frac = 0

		self._set_config_type("bpp_rx", TYPE_INT, "1000")
		self._set_config_type("bpp_tx", TYPE_INT, "1000")
		
		self._set_config_type("col_sky_s", TYPE_STRING, "#0099ff")
		self._set_config_type("col_sky_e", TYPE_STRING, "#ff2200")
		self._set_config_type("col_water_s", TYPE_STRING, "#0077aa")
		self._set_config_type("col_water_e", TYPE_STRING, "#aa2200")
		self._set_config_type("col_rx", TYPE_STRING, "#698287")
		self._set_config_type("col_tx", TYPE_STRING, "#876969")

		self._add_timer(75, self._do_gfx)
		self._add_thread(self._stats_thread)

	def _do_gfx(self):
		o = self._new_output()

		# 
		# Update the state of the display
		#
		
		# The colour used to draw the sky
		skycol_s = self._get_config("col_sky_s")
		skycol_e = self._get_config("col_sky_e")
		skycol = colour_trans(skycol_s, skycol_e, self.swap_frac)
		
		# The colour used to draw the water
		watercol_s = self._get_config("col_water_s")
		watercol_e = self._get_config("col_water_e")
		watercol = colour_trans(watercol_s, watercol_e, self.swap_frac)
		
		# Create bubbles based on CPU usage
		for n in range(2):
			if uniform(0,100) <= self.cpu:
				self.water.create_bubble()

		self.water.update()
		self.duck.update(self.water)

		# Calculate the score and number of active fishies
		arx = 0
		atx = 0
		for fishy in self.tx_fishies[:]:
			if fishy.is_offscreen():
				self.tx_fishies.remove(fishy)
			else:
				atx = atx + fishy.score()

		for fishy in self.rx_fishies[:]:
			if fishy.is_offscreen():
				self.rx_fishies.remove(fishy)
			else:
				arx = arx + fishy.score()

		# If the score of fishies on display is greater than the
		# network score, then slow the fishies down. 	
		#
		# Otherwise, use the points to speed up the fishies on display.
		# The unused points are used to spawn new fishies.

		nrx = len(self.rx_fishies)
		drx = 0
		if self.rx_points < arx:
			drx = -0.5
		elif nrx:
			drx = min(5, float(self.rx_points) / nrx)
		# Else there aren't any fishies
		for fishy in self.rx_fishies:
			arx = arx + fishy.accel(drx)
			fishy.update(self.water)

		# Now do the same for Tx:
		ntx = len(self.tx_fishies)
		dtx = 0
		if self.tx_points < atx:
			dtx = -0.5
		elif ntx:
			dtx = min(5, float(self.tx_points) / ntx)
		# Else there aren't any fishies to speed up
		for fishy in self.tx_fishies:
			atx = atx + fishy.accel(dtx)
			fishy.update(self.water)

		# Any tx/rx points left over can be used to make new fishies.

		tx_left = max(0, self.tx_points - atx)
		rx_left = max(0, self.rx_points - arx)
		if tx_left > 0 and tx_left >= 20*ntx and ntx < MAX_FISHIES:
			fishcolour = self._get_config("col_tx")
			self.tx_fishies.append(Fishy(uniform(self.water.height, CANVAS_HEIGHT), tx_left, 1, fishcolour))
		if rx_left > 0 and rx_left >= 20*nrx and nrx < MAX_FISHIES: 
			fishcolour = self._get_config("col_rx")
			self.rx_fishies.append(Fishy(uniform(self.water.height, CANVAS_HEIGHT), rx_left, -1, fishcolour))

		# 
		# Render the scene
		#
		
		# Sky
		self._gfx = "clear, color %s, polygon -1 -1 -1 1 1 1 1 -1 1," % skycol
		# Duck
		self._gfx = self._gfx + self.duck.render() + ","
		# Water
		self._gfx = self._gfx + self.water.render(watercol) 
		# Fish
		for fishy in self.rx_fishies: self._gfx = self._gfx + "," + fishy.render()
		for fishy in self.tx_fishies: self._gfx = self._gfx + "," + fishy.render()

		# 
		# Output the graphics to the display
		#
		
		o.set("gfx", self._gfx)
		self._send_output(o)

		self._add_timer(50, self._do_gfx)

	def get_configurator(self):
		c = self._new_configurator()
		c.set_name("Bubblefishymon")
		
		c.add_title("Visual settings")
		c.add_color_selector("Sky colour (No swap)", "col_sky_s", "The colour to draw the sky when no swap is used.")
		c.add_color_selector("Sky colour (Full swap)", "col_sky_e", "The colour to draw the sky when all swap is used.")
		c.add_color_selector("Water colour (No swap)", "col_water_s", "The colour to draw the water when no swap is used.")
		c.add_color_selector("Water colour (Full swap)", "col_water_e", "The colour to draw the water when all swap is used.")
		c.add_color_selector("Rx fishy colour", "col_rx", "The colour to draw fish that represent received network traffic")
		c.add_color_selector("Tx fishy colour", "col_tx", "The colour to draw fish that represent transmitted network traffic")
		
		c.add_title("Behavioural settings")
		c.add_spin("Rx Bytes Per Point", "bpp_rx", "Defines how many bytes there are in 1 network activity point for received traffic. These points are used to determine how fast the fish move, and whether to create new fish.", 10, 100000)
		c.add_spin("Tx Bytes Per Point", "bpp_tx", "Defines how many bytes there are in 1 network activity point for received traffic. These points are used to determine how fast the fish move, and whether to create new fish.", 10, 100000)

		return c

	def _stats_thread(self):
		while not self._is_stopped():
			self._bytes_per_point_rx = self._get_config("bpp_rx")
			self._bytes_per_point_tx = self._get_config("bpp_tx")
		
			rx_spd,tx_spd = libdesklets.net.get_speed_in_out("eth0")
			ram_t,ram_u = libdesklets.memory.get_ram()
			swap_t,swap_u = libdesklets.memory.get_swap()
			cpu = libdesklets.cpu.get_load()[1]

			self.rx_points = float(rx_spd) / self._bytes_per_point_rx
			self.tx_points = float(tx_spd) / self._bytes_per_point_tx

			self.water.height = (1 - float(ram_u) / ram_t) * CANVAS_HEIGHT
			self.swap_frac = float(swap_u) / swap_t

			self.cpu = cpu

			time.sleep(0.075)
		
		
def new_sensor(args): return BubbleFishyMon(*args)
			
