from sensor.Sensor import Sensor
from utils.datatypes import *
from utils import i18n

import time
import math
import commands

class TimeSensor(Sensor):
	"""
	Sensor for showing time. This is a fork of the Clock sensor by Martin Grimme
	<martin@pycage.de> to add uptime functionality, and make it more flexible with regards to 
	12-hour time format.

	Configuration data:

	- show-second-hand : Whether to show the second hand on the analogue clock
    - show-seconds     : Whether to show seconds on the digital clock
    - show-date        : Whether to show the date
    - date-format      : Whether to use the long or short date format ("short" or "long")
    - show-uptime      : Whether to show the uptime
    - uptime-format    : Format of uptime: %d, %h, %m and %s = days, hours, minutes, seconds
    - mode             : Whether to use a 12 hour or 24 hour clock ("12hr" or "24hr")
    - timezone         : The timezone to use

    Output data:

    - hands            : Drawing instructions for hands of analogue clock
    - hours            : Hours part of digital clock
    - minutes          : Minutes part of digital clock
    - seconds          : Seconds part of digital clock (blank if show-seconds is false)
    - daytime          : am or pm (blank if mode is 24hr)
    - time             : Time as hh:mm[:ss], formatted according to mode and show-seconds
    - date             : Date according to date-format, blank if show-date is false
    - uptime           : Uptime, blank if show-uptime is false

    - show-seconds     : Value of show-seconds configuration parameter
    - show-date        : Value of show-date configuration parameter
    - show-uptime      : Value of show-uptime configuration parameter
	"""

	# the default clock hands
	HAND_HOUR = [-0.03, -0.2,
	                0.03, -0.2,
	                0.03, 0.6,
	                -0.03, 0.6]

	HAND_MINUTE = [-0.03, -0.2,
	                0.03, -0.2,
	                0.03, 0.9,
	                -0.03, 0.9]

	HAND_SECOND = [-0.02, -0.2,
	                0.02, -0.2,
	                0.02, 1.0,
	                -0.02, 1.0]

	MODE_12H = "12h"
	MODE_24H = "24h"

	DATE_LONG  = "long"
	DATE_SHORT = "short"

	def __init__(self, show_seconds = 0, show_second_hand = 1, show_date = 1, show_uptime = 0, 
                    mode = "24h", date_format = "long", uptime_format="%dd %hh %mm", 
	                hand_hour = "", hand_minute = "", hand_second = ""):
		"""
		Initalise. 
		    - show_seconds     : Whether to show seconds on the digital clock
		    - show_second_hand : Whether to show the second hand on the analogue clock
		    - show_date        : Whether to show the date
		    - show_uptime      : Whether to show uptime
			- mode             : 12h or 24h clock
			- date_format      : Whether to show long or short dates by default
		    - uptime_format    : The uptime format to use by default
		    - hand_*           : 8-element arrays specifying the hands of the clock.
		"""

		global _; _ = i18n.Translator("clock-sensor")

		self.__hand_hour = hand_hour or self.HAND_HOUR
		self.__hand_minute = hand_minute or self.HAND_MINUTE
		self.__hand_second = hand_hour or self.HAND_SECOND

		self.__timezones = [(_("local"), "localtime"),
		                    (_("UTC"), "UTC0")]

		self.__timezone = ""
		self.__offset = 0

		Sensor.__init__(self)
		self._set_config_type("show-second-hand", TYPE_BOOL, show_second_hand)
		self._set_config_type("show-seconds", TYPE_BOOL, show_seconds)
		self._set_config_type("show-date", TYPE_BOOL, show_date)
		self._set_config_type("date-format", TYPE_STRING, date_format)
		self._set_config_type("show-uptime", TYPE_BOOL, show_uptime)
		self._set_config_type("uptime-format", TYPE_STRING, uptime_format)
		self._set_config_type("mode", TYPE_STRING, mode)
		self._set_config_type("timezone", TYPE_STRING, "localtime")

		self.__timezones += self.__load_timezones()

		self._add_timer(1000, self.__get_data)

	def __load_timezones(self):
		"Load the timezones from timezones.dat and return tuples (name, tzname)."

		zones = []
		try:
			fd = open("timezones.dat", "r")
			try:
				lines = fd.readlines()
			finally:
				fd.close()
			for l in lines:
				name = l.strip()
				zones.append((name, name))
			#end for
		except:
			pass

		return zones

	def get_configurator(self):
		"Get the configurator"

		configurator = self._new_configurator()
		configurator.set_name(_("Clock"))

		configurator.add_title(_("Time"))
		configurator.add_checkbox(_("Show second hand"), "show-second-hand",
													_("Toggles the second hand on and off"))
		configurator.add_checkbox(_("Show seconds"), "show-seconds",
													_("Toggles the seconds on the digital clock on and off"))
		configurator.add_option(_("Display mode (digital clock)"), "mode",
														_("12h or 24h display mode"),
														[(_("12h"), self.MODE_12H),
														 (_("24h"), self.MODE_24H)])
		configurator.add_option(_("Timezone"), "timezone",
														_("The timezone to use"),
														self.__timezones)

		configurator.add_title(_("Date"))
		configurator.add_checkbox(_("Show date"), "show-date",
													_("Toggles the date on and off"))
		configurator.add_option(_("Date format"), "date-format", 
								_("The date format to use"),
								[(_("Long"), self.DATE_LONG),
								 (_("Short"), self.DATE_SHORT)])

		configurator.add_title(_("Uptime"))
		configurator.add_checkbox(_("Show uptime"), "show-uptime",
													_("Toggles the uptime on and off"))
		configurator.add_entry(_("Uptime format"), "uptime-format",
													_("Days:\t%d\nHours:\t%h\nMinutes:\t%m\nSeconds:\t%s"))

		return configurator

	def __rotate(self, x, y, angle):
		"Rotate the point by the given angle"

		cos_a = math.cos(angle)
		sin_a = math.sin(angle)
		rx = x * cos_a - y * sin_a
		ry = y * cos_a + x * sin_a

		return (rx, ry)


	def __do_hand(self, points, angle):
		"Generate and return the drawing instructions for a clock hand"

		p = ""
		for i in xrange(0, len(points), 2):
			x = points[i]
			y = points[i + 1]
			rx, ry = self.__rotate(x, y, angle)
			p += " %(rx)f %(ry)f" % vars()
		
		return p

	def __get_data(self):
		"Send the clock data"
		
		# get the timezone offset in seconds
		timezone = self._get_config("timezone")
		if (timezone != self.__timezone):
			self.__timezone = timezone
			offset = commands.getoutput("TZ=\"%(timezone)s\" date +%%z" % vars())
			sign = (offset[0] == "+") and 1 or -1
			hours = int(offset[1:3])
			mins = int(offset[3:5])
			self.__offset = sign * (mins * 60 + hours * 3600)
		
		show_second_hand = self._get_config("show-second-hand")
		show_secs = self._get_config("show-seconds")
		show_date = self._get_config("show-date")
		date_format = self._get_config("date-format")
		mode = self._get_config("mode")
		show_uptime = self._get_config("show-uptime")
		
		# get the current time for the selected timezone
		current_time = time.gmtime(time.time() + self.__offset)
		
		hour = current_time[3]
		minute = current_time[4]
		second = current_time[5]
		
		h_angle = -(2 * math.pi) / 12 * (hour + (minute / 60.0))
		m_angle = -(2 * math.pi) / 60 * minute
		s_angle = -(2 * math.pi) / 60 * second
		
		gfx_hour = self.__do_hand(self.__hand_hour, h_angle)
		gfx_minute = self.__do_hand(self.__hand_minute, m_angle)
		
		if show_second_hand:
			gfx_second = self.__do_hand(self.__hand_second, s_angle)
		else:
			gfx_second = ""
		
		
		if mode == self.MODE_12H:
			if hour > 12:
				hour -= 12
				daytime = "pm"
			else:
				daytime = "am"
		else:
			daytime = ""
		
		signal = self._new_output()
		
		# digital clock
		signal.set("hours", "%(hour)02d" % vars())
		signal.set("minutes", "%(minute)02d" % vars())
		
		if show_secs:
			signal.set("seconds", "%(second)02d" % vars())
		else:
			signal.set("seconds", "")
		
		signal.set("daytime", daytime)
		
		signal.set("time", self.__format_time(current_time, mode, show_secs))
		
		# date
		if show_date:
			signal.set("date", self.__get_date(current_time, date_format))
		else:
			signal.set("date", "")
		
		
		# analogue clock
		signal.set("hands", "clear, color black, "
		                    "polygon %(gfx_hour)s 1, "
		                    "polygon %(gfx_minute)s 1,"
		                    "color red,"
		                    "polygon %(gfx_second)s 1" % vars())
		
		# uptime
		if show_uptime:
			signal.set("uptime", self.__get_uptime())
		else:
			signal.set("uptime", "")
		
		# toggles (useful for hiding elements of the display)
		signal.set("show-seconds", show_secs)
		signal.set("show-date", show_date)
		signal.set("show-uptime", show_uptime)
		
		self._send_output(signal)
		return 1

	def __get_date(self, current_time, date_format):
		"Get the date string"
	
		if date_format == self.DATE_LONG:
			return time.strftime("%a %d %b %y", current_time)
		else:
			return time.strftime("%x", current_time)

	def __get_uptime(self):
		"Get the current uptime"

		# Stolen from LTVUptime sensor
		data = commands.getoutput("cat /proc/uptime")

		uptime = float( data.split()[0] )
		days = int( uptime / 60 / 60 / 24 )
		uptime = uptime - days * 60 * 60 * 24
		hours = int( uptime / 60 / 60 )
		uptime = uptime - hours * 60 * 60
		minutes = int( uptime / 60 )
		uptime = uptime - minutes * 60
		seconds = int( uptime )

		time = self._get_config("uptime-format")
		time = time.replace("%d", str(days))
		time = time.replace("%h", str(hours))
		time = time.replace("%m", str(minutes))
		time = time.replace("%s", str(seconds))

		return time

	def __format_time(self, current_time, mode, show_secs):
		"Format time according to mode"

		fmt = ""
		if mode == self.MODE_12H:
			fmt += "%I"
		else:
			fmt += "%H"
		
		fmt += ":%M"
		
		if show_secs:
			fmt += ":%S"
		
		if mode == self.MODE_12H:
			fmt += " %p"
		
		return time.strftime(fmt, current_time)

def new_sensor(args): return apply(TimeSensor, args)
