
from utils.datatypes import *
from sensor.Sensor import Sensor

import commands
import re

class SpeedfreqMon (Sensor):
	"""
	Sensor that periodically calls the 'speedfreq' command to find the current, min and max CPU speeds.
	Requires that speedfreqd is running, which in turn requires a 2.6 kernel with Intel Enhanced SpeedStep
	(Centrino) support.
		
	Output data:
		- min				 : The minimum CPU speed, in MHz
		- max			 : The maximum CPU speed, in MHz
		- current		 : The current CPU speed, in MHZ
		- percentage	 : The percentage as floating point number of max CPU speed used
		- idle				 : The percentage of the CPU that is idle
		- policy			 : The policy in use
		- error			 : True if the sensor could not get speedfreq information
	"""

	def __init__(self, interval=1000):
		
		Sensor.__init__(self)		
		self._add_timer (int (interval), self.update_display)

	def update_display (self):
	
		(min_speed, max_speed, current_speed, idle, policy, error) = self.get_speeds ()
	
		data = self._new_output()
	
		percentage = 0
		
		if not error:
			percentage = float (current_speed) / max_speed * 100
	
		data.set ("min", "%dMHz" % (min_speed))
		data.set ("max", "%dMHz" % (max_speed))
		data.set ("current", "%dMHz" % (current_speed))
		data.set ("percentage", percentage)
		data.set ("idle", "%.2f%% idle" % (idle))
		data.set ("policy", policy)
		data.set ("error", error)
		
		self._send_output(data)

		return 1

	def get_speeds (self):
		"""
		Exectute "speedfreq -sc" and get the results.
		"""
		
		min_speed = 0
		max_speed = 0
		current_speed = 0
		idle = 0
		policy = ""
		error = 1
		
		output = commands.getoutput ('speedfreq -sc')
		
		speed_match = re.search ('CPU speed: min (.*)MHz, max (.*)MHz, current (.*)MHz; (.*)% idle\n',
								 output, re.I)

		policy_match = re.search ('Current policy is (.*)\n', output, re.I)
		
		if speed_match and policy_match:
			min_speed = int (speed_match.group (1))
			max_speed = int (speed_match.group (2))
			current_speed = int (speed_match.group (3))
			idle = float (speed_match.group (4))
			policy = policy_match.group (1)
			error = 0
			
		return (min_speed, max_speed, current_speed, idle, policy, error)
				
def new_sensor(args):
        return apply(SpeedfreqMon, args)
