# vim: expandtab : tabstop=4 : shiftwidth=4 : smarttab
# coding: UTF-8
# TODO:
# - Currently, sensors are splitted into 3 groups. Once gdesklets supports
#   un-numbered arrays, one sensor will be enough for all 3 groups.
# - The non-Psi versions don't work anymore with the Psi version.
# - read min/max values, and color accordingl
# Version 0.8
# - only updateed displays
# Version 0.7
# - optionally splitted it into 3 sensors: temperature/voltage/fan
# Version 0.6
# - parse output of 'sensors' instead of reading /proc/... directly
# Version 0.5
# Version 0.4
# Version 0.3
# Version 0.2
# Version 0.1
# Originally written by Elias Pschernig <elias@users.sf.net>.
# Originally adapted from code posted at
# http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=207315.
# Originally based on the hddtemp sensor by Sebastien Bacher.

# command to parse
command = "sensors -A"
# downscaling of values for plotters
sens_temp = 0.1    # temperatures
sens_volt = 0.005  # voltages
sens_fan = 2       # fans

type_names = {
    "temp" : "Temperature",
    "volt" : "Voltage",
    "fan"  : "Fan Rotation",
    ""     : "all"}

from sensor.Sensor import Sensor
from utils.datatypes import *
from utils import i18n

import time
import sys
import re
import commands

class LMSensorsSensor(Sensor):

    """Sensor for retrieving lm-sensors information."""

    def __init__(self, args = ""):
        """Sets up the sensor."""
        global _; _ = i18n.Translator("lmsensors-sensor")
        Sensor.__init__(self)
        # sensorlist keys: 
        # "name"         Configuration name.
        # "label"        Label to display.
        # "sensitivity"  Plotter sensitivity.
        # "value"        Sensor value.
        # "plot"         Plotter value.
        # "last"         Previous value.
        # "show"         Display this sensor?
        self.__sensorlist = []
        self.__active = 0
        self.__type = args

        self._set_config_type("font", TYPE_STRING, "Sans 10")
        self._set_config_type("color", TYPE_STRING, "Black")

        self._watch_config (self.__config)

        # It's not allowed to send output from inside this function!
        # It's also not allowed to send output from another thread(!), so we
        # need a ticker.
        self._add_timer(1000, self.__tick)
        self._add_thread(self.__get_value_thread)

    def get_configurator(self):
        """Called everytime the config menu is selected."""

        type_name = type_names[self.__type]

        configurator = self._new_configurator()

        configurator.add_title(_("LMSensors"))
        configurator.add_spin(_("Update every (seconds):"), "update",
            _("Updates the sensor value every x seconds"), 1, 900)

        configurator.add_font_selector(_("Font:"),
            "font", _("Font used for %s display." % type_name))
        configurator.add_color_selector(_("Color:"),
            "color", _("Color used for %s display." % type_name))

        configurator.set_name(type_name)
        if self.__active:
            if len (self.__sensorlist):
                configurator.add_title(_("Available %s sensors:" % 
                    type_name))
                for sensor in self.__sensorlist: 
                    configurator.add_checkbox (sensor["name"], sensor["name"],
                        _("Toggle this sensor"))
            else:
                configurator.add_title(_("No %s sensors were found." %
                     type_name))
                configurator.add_title(_("Is lmsensors installed and working?"))
        else:
            configurator.add_title(
                _("Not done reading hardware sensors yet..."))

        return configurator

    def __get_sensor_data (self, append = False):
        """Runs the sensor program and parses the output."""
        (status, output) = commands.getstatusoutput (command)

        if status != 0: return

        lines = output.splitlines()
 
        for line in lines: 
            line = unicode(line)
            # line with just one single word on it -> chip name
            if re.match ("^[^:\s]+$", line):
                chipname = line
            else:
                # name colon number open-parenthesis
                # TODO: also extract max and min, and color accordingly
                val = re.match ("^\s*([^:]+?)\s*:\s*([^\(]+?)\s*\(", line)
                if val:
                    val = val.groups()
                    if val[0] != "alarms":
                        sensor = {}
                        sensor["name"] = re.sub ("[^-\w\d\/]", "_", chipname + "/" + val[0])
                        sensor["value"] = re.sub ("\xb0", "°", val[1])
                        v = float (re.sub ("[^\d]+$", "", sensor["value"]))
                        if append:
                            # new entry
                            is_RPM = sensor["value"].find ("RPM") != -1
                            is_V = sensor["value"].find ("V") != -1
                            sensor["label"] = val[0]
                            sensor["plot"] = 50
                            sensor["last"] = v
                            if is_RPM:
                                type = "fan"
                                sensor["sensibility"] = sens_fan
                            elif is_V:
                                type = "volt"
                                sensor["sensibility"] = sens_volt
                            else:
                                type = "temp"
                                sensor["sensibility"] = sens_temp
                            if type == self.__type or self.__type == "":
                                self.__sensorlist.append (sensor)
                                self._set_config_type(sensor["name"],
                                    TYPE_BOOL, 1)
                        else:
                            # update value
                            for s in self.__sensorlist:
                                if s["name"] == sensor["name"]:
                                    s["value"] = sensor["value"] 
                                    s["plot"] += (v - s["last"]) / s["sensibility"]
                                    s["last"] = v
        self.__updated = True
        return

    def __output_info (self):
        """Outputs the stored info."""
        output = self._new_output()

        i = 0
        for sensor in self.__sensorlist:
            if sensor["show"]:
                output.set (self.__type + "_label[%i]" % i, sensor["label"])
                output.set (self.__type + "_value[%i]" % i, sensor["value"])
                output.set ("font[%i]" % i, self._get_config ("font"))
                output.set ("color[%i]" % i, self._get_config ("color"))

                if self.__updated:
                    # Only update the plotters when new values have been read.
                    output.set (self.__type + "_plot[%i]" % i, sensor["plot"]);

                i = i + 1
            #end if
        #end for
        if i:
            output.set (self.__type + "_count", i);

        self._send_output(output)
        self.__updated = False

    def __config (self, name = "", value = ""):
        for sensor in self.__sensorlist:
            sensor["show"] = self._get_config (sensor["name"])

    def __tick (self):
        """Executed once a second inside the main thread."""

        if self.__active:
            self.__output_info ()  

        return True

    def __get_value_thread (self):
        """The work thread."""
        print "Lmsensors %s sensor starting." % type_names[self.__type]
        self._set_config_type("update", TYPE_INT, 10)
        self.__get_sensor_data (append = True)
        self.__config ()
        self.__active = 1
        while (not self._is_stopped()):
            time.sleep (self._get_config ("update"))
            self.__get_sensor_data ()

def new_sensor (args): return apply (LMSensorsSensor, args)
